from enum import Enum
from typing import NewType, NamedTuple, Optional, Mapping, Sequence, Any

import typeit
from inflection import camelize
from typeit.schema import Invalid


class ContentTypeFormat(Enum):
    JSON = 'application/json'
    XML = 'application/xml'
    TEXT = 'text/plain'
    FORM_URLENCODED = 'application/x-www-form-urlencoded'
    BINARY_STREAM = 'application/octet-stream'
    EVENT_STREAM = 'text/event-stream'
    """ server-side events
    """
    ANYTHING = '*/*'


MediaTypeCharset = NewType('MediaTypeCharset', str)


class ContentTypeTag(NamedTuple):
    format: ContentTypeFormat
    charset: Optional[MediaTypeCharset]


class ContentTypeTagSchema(typeit.schema.primitives.Str):
    def deserialize(self, node, cstruct: str) -> ContentTypeTag:
        """ Converts input string value ``cstruct`` to ``ContentTypeTag``
        """
        try:
            tag_str = super().deserialize(node, cstruct)
        except Invalid as e:
            error = Invalid(node, "Media Type should be a string", cstruct)
            error.add(e)
            raise error

        media_format, *param = tag_str.split(';')
        try:
            typed_format = ContentTypeFormat(media_format)
        except ValueError:
            raise Invalid(node, f"Unsupported Media Type format: {media_format}", media_format)

        if param:
            param = [x for x in param[0].split('charset=') if x.strip()]
            if param:
                charset = param[0]
            else:
                charset = None
        else:
            charset = None
        return ContentTypeTag(
            format=typed_format,
            charset=charset
        )

    def serialize(self, node, appstruct: ContentTypeTag) -> str:
        """ Converts ``ContentTypeTag`` back to string value suitable for JSON/YAML
        """
        rv = [appstruct.format.value]
        if appstruct.charset:
            rv.extend([';', "charset=", appstruct.charset])

        return super().serialize(node, ''.join(rv))


class RefTo(Enum):
    SCHEMAS = '#/components/schemas/'
    LINKS = '#/components/links/'


class Ref(NamedTuple):
    location: RefTo
    name: str


class RefSchema(typeit.schema.primitives.Str):
    REF_PREFIX: Sequence[str] = ['#', 'components']
    REF_LOCATIONS: Mapping[str, RefTo] = {
        'schemas': RefTo.SCHEMAS,
        'links': RefTo.LINKS,
    }

    def deserialize(self, node, cstruct: str) -> Ref:
        """ Converts input string value ``cstruct`` to ``Ref``
        """
        try:
            ref_str = super().deserialize(node, cstruct)
        except Invalid as e:
            error = Invalid(node, "Reference should be a string", cstruct)
            error.add(e)
            raise error

        try:
            *prefix, location, ref_name = ref_str.split('/')
        except (ValueError, AttributeError):
            raise Invalid(node, "Invalid reference format", ref_str)

        if prefix != self.REF_PREFIX:
            raise Invalid(node, f"Reference is not prefixed with {'/'.join(self.REF_PREFIX)}", ref_str)

        try:
            ref_location = self.REF_LOCATIONS[location]
        except KeyError:
            raise Invalid(node, "Unrecognised reference location", ref_str)

        return Ref(ref_location, ref_name)

    def serialize(self, node, appstruct: Ref) -> str:
        """ Converts ``Ref`` back to string value suitable for JSON/YAML
        """
        rv = [appstruct.location.value, appstruct.name]
        return super().serialize(node, ''.join(rv))


class EmptyValue(NamedTuple):
    """ Sometimes spec contains schemas like:
    {
        "type": "array",
        "items": {}
    }

    In that case we need a strict type that would check that its serialized representation
    exactly matches the empty schema value {}. This object serves that purpose.
    """
    pass


_empty = EmptyValue()


class EmptyValueSchema(typeit.schema.meta.SchemaType):
    def deserialize(self, node, cstruct: Any) -> EmptyValue:
        """ Converts input string value ``cstruct`` to ``EmptyValue``
        """
        if cstruct == {}:
            error = Invalid(node, "Not an empty type", cstruct)
            raise error
        return _empty

    def serialize(self, node, appstruct: EmptyValue) -> Mapping[Any, Any]:
        """ Converts ``EmptyValue`` back to a value suitable for JSON/YAML
        """
        return {}


TypeGenerator = (typeit.TypeConstructor
                 & ContentTypeTagSchema[ContentTypeTag]  # type: ignore
                 & RefSchema[Ref]                        # type: ignore
                 & EmptyValueSchema[EmptyValue]          # type: ignore
                 & typeit.flags.GlobalNameOverride(lambda x: camelize(x, uppercase_first_letter=False))
                 )



import argparse
import sys

from pkg_resources import get_distribution

from . import check
from ..info import DISTRIBUTION_NAME


def main(args=None, in_channel=sys.stdin, out_channel=sys.stdout):
    parser = argparse.ArgumentParser(description='OpenAPI Type')
    parser.add_argument('-V', '--version', action='version',
                        version=f'{DISTRIBUTION_NAME} {get_distribution(DISTRIBUTION_NAME).version}')
    subparsers = parser.add_subparsers(title='sub-commands',
                                       description='valid sub-commands',
                                       help='additional help',
                                       dest='sub-command')
    # make subparsers required (see http://stackoverflow.com/a/23354355/458106)
    subparsers.required = True

    # $ <cmd> gen
    # ---------------------------
    check.setup(subparsers)

    # Parse arguments and config
    # --------------------------
    if args is None:
        args = sys.argv[1:]
    args = parser.parse_args(args)

    # Set up and run
    # --------------
    args.run_cmd(args, in_channel=in_channel, out_channel=out_channel)



from enum import Enum

from typing import Literal, NewType
from typing import Mapping
from typing import Any, NamedTuple, Optional, Sequence, FrozenSet, Union

from pyrsistent import pmap, pvector
from pyrsistent.typing import PVector, PMap

from .custom_types import TypeGenerator, ContentTypeTag, Ref, EmptyValue


__all__ = ('parse_spec', 'serialize_spec', 'OpenAPI')


class IntegerValue(NamedTuple):
    type: Literal['integer']
    format: str = ''
    example: Optional[int] = None
    default: Optional[int] = None
    minimum: Optional[int] = None
    maximum: Optional[int] = None


class FloatValue(NamedTuple):
    type: Literal['number']
    format: str = ''
    example: Optional[float] = None
    default: Optional[float] = None


class StringValue(NamedTuple):
    type: Literal['string']
    format: str = ''
    description: str = ''
    enum: PVector[str] = pvector()
    default: Optional[str] = None
    pattern: Optional[str] = None
    example: str = ''


class BooleanValue(NamedTuple):
    type: Literal['boolean']
    default: Optional[bool] = None


class Reference(NamedTuple):
    ref: Ref


RecursiveAttrs = Mapping[str, 'SchemaType']  # type: ignore


class ObjectWithAdditionalProperties(NamedTuple):
    type: Literal['object']
    additional_properties: Mapping[str, Any]


class ArrayValue(NamedTuple):
    type: Literal['array']
    items: 'SchemaType'  # type: ignore


class ObjectValue(NamedTuple):
    type: Literal['object']
    properties: RecursiveAttrs
    required: FrozenSet[str] = frozenset()
    description: str = ''
    xml: Mapping[str, Any] = pmap()


class InlinedObjectValue(NamedTuple):
    properties: RecursiveAttrs
    required: FrozenSet[str]
    description: str = ''


class ResponseRef(NamedTuple):
    """ Values that are referenced as $response.body#/some/path
    """
    operation_id: str
    parameters: Mapping[str, str]


class ObjectRef(NamedTuple):
    """ Values that are referenced as #/components/schemas/<SomeType>
    """
    ref: str


class ProductSchemaType(NamedTuple):
    all_of: Sequence['SchemaType']  # type: ignore


class UnionSchemaTypeAny(NamedTuple):
    any_of: Sequence['SchemaType']  # type: ignore


class UnionSchemaTypeOne(NamedTuple):
    one_of: Sequence['SchemaType']  # type: ignore


SchemaType = Union[ StringValue  # type: ignore
                  , IntegerValue
                  , FloatValue
                  , BooleanValue
                  , ObjectValue
                  , ArrayValue
                  , ResponseRef
                  , Reference
                  , ProductSchemaType
                  , UnionSchemaTypeAny
                  , UnionSchemaTypeOne
                  , ObjectWithAdditionalProperties
                  , InlinedObjectValue
                  , EmptyValue
                  ]


class Components(NamedTuple):
    schemas: Mapping[str, SchemaType]
    links: Mapping[str, SchemaType] = pmap()
    request_bodies: Mapping[str, Any] = pmap()
    security_schemes: Mapping[str, Any] = pmap()


class ServerVar(NamedTuple):
    default: str
    enum: Sequence[str]
    description: str = ''


class Server(NamedTuple):
    url: str
    description: str = ''
    variables: Mapping[str, ServerVar] = pmap()


class InfoLicense(NamedTuple):
    name: str
    url: str = ''


class InfoContact(NamedTuple):
    name: Optional[str]
    email: Optional[str]
    url: Optional[str]


class Info(NamedTuple):
    version: str
    """ API version
    """
    title: str
    license: Optional[InfoLicense]
    contact: Optional[InfoContact]
    terms_of_service: str = ''
    description: str = ''


class SpecFormat(Enum):
    V3_0_0 = '3.0.0'
    V3_0_1 = '3.0.1'
    V3_0_2 = '3.0.2'


class ParamLocation(Enum):
    QUERY = 'query'
    HEADER = 'header'
    PATH = 'path'
    COOKIE = 'cookie'


class ParamStyle(Enum):
    """
    * https://swagger.io/specification/#style-values
    * https://swagger.io/specification/#style-examples
    """
    FORM = 'form'
    SIMPLE = 'simple'
    MATRIX = 'matrix'
    LABEL = 'label'
    SPACE_DELIMITED = 'spaceDelimited'
    PIPE_DELIMITED = 'pipeDelimited'
    DEEP_OBJECT = 'deepObject'


class OperationParameter(NamedTuple):
    name: str
    in_: ParamLocation
    schema: SchemaType
    required: bool = False
    description: str = ''
    style: Optional[ParamStyle] = None
    explode: Optional[bool] = None


HTTPCode = NewType('HTTPCode', str)
HeaderName = NewType('HeaderName', str)


class Header(NamedTuple):
    """ response header
    """
    schema: SchemaType
    description: str = ''


class MediaType(NamedTuple):
    """ https://swagger.io/specification/#media-type-object
    """
    schema: Optional[SchemaType] = None
    example: Union[None, str, PMap[str, Any]] = None
    examples: Mapping[str, Any] = pmap()
    encoding: Mapping[str, Any] = pmap()


class Response(NamedTuple):
    """ Response of an endpoint
    """
    content: PMap[ContentTypeTag, MediaType] = pmap()
    headers: PMap[HeaderName, Header] = pmap()
    description: str = ''


class ExternalDoc(NamedTuple):
    url: str
    description: str = ''


class RequestBody(NamedTuple):
    """ https://swagger.io/specification/#request-body-object
    """
    content: Mapping[ContentTypeTag, Any]
    description: str = ''
    required: bool = False


class Operation(NamedTuple):
    """ https://swagger.io/specification/#operation-object
    """
    responses: Mapping[HTTPCode, Response]
    external_docs: Optional[ExternalDoc]
    summary: str = ''
    operation_id: str = ''
    parameters: FrozenSet[Union[OperationParameter, Reference]] = frozenset()
    request_body: Union[None, RequestBody, Reference] = None
    description: str = ''
    tags: FrozenSet[str] = frozenset()
    callbacks: Mapping[str, Mapping[str, Any]] = pmap()
    security: Optional[Any] = None


class PathItem(NamedTuple):
    """ Describes endpoint methods
    """
    head: Optional[Operation]
    get: Optional[Operation]
    post: Optional[Operation]
    put: Optional[Operation]
    patch: Optional[Operation]
    delete: Optional[Operation]
    trace: Optional[Operation]
    servers: Sequence[Server] = pvector()
    ref: Optional[Ref] = None
    summary: str = ''
    description: str = ''


SecurityName = NewType('SecurityName', str)


class SpecTag(NamedTuple):
    name: str
    external_docs: Optional[ExternalDoc]
    description: str = ''


class OpenAPI(NamedTuple):
    openapi: SpecFormat
    """ Spec format version
    """
    info: Info
    """ Various metadata
    """
    paths: Mapping[str, PathItem]
    components: Components = Components(schemas=pmap(), links=pmap())
    servers: Sequence[Server] = pvector()
    security: Sequence[Mapping[SecurityName, Sequence[str]]] = pvector()
    tags: Sequence[SpecTag] = pvector()
    external_docs: Optional[ExternalDoc] = None


overrides = {
    OperationParameter.in_: 'in',
    Reference.ref: '$ref',
    PathItem.ref: '$ref',
}


parse_spec, serialize_spec = TypeGenerator & overrides ^ OpenAPI



from pathlib import Path
from setuptools import setup
from setuptools import find_packages


here = Path(__file__).absolute().parent


EXTRAS = frozenset({
    'third_party',
})


def extras_require(all_extras=EXTRAS):
    """ Get map of all extra requirements
    """
    return {
        x: requirements(here / 'requirements' / 'extras' / f'{x}.txt') for x in all_extras
    }


def requirements(at_path: Path):
    with at_path.open() as f:
        rows = f.read().strip().split('\n')
        requires = []
        for row in rows:
            row = row.strip()
            if row and not (row.startswith('#') or row.startswith('http')):
                requires.append(row)
    return requires


with (here / 'README.rst').open() as f:
    README = f.read()


# Setup
# ----------------------------

setup(name='openapi-type',
      version='0.0.13',
      description='OpenAPI Type',
      long_description=README,
      classifiers=[
          'Intended Audience :: Developers',
          'License :: OSI Approved',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Operating System :: POSIX',
      ],
      author='Maxim Avanov',
      author_email='maxim.avanov@gmail.com',
      url='https://github.com/avanov/openapi-type',
      keywords='typing json yaml openapi oas swagger schema serialization deserialization structured-data',
      packages=find_packages(exclude=['tests', 'tests.*']),
      include_package_data=True,
      zip_safe=False,
      test_suite='tests',
      tests_require=['pytest', 'coverage'],
      install_requires=requirements(here / 'requirements' / 'minimal.txt'),
      extras_require=extras_require(),
      entry_points={
          'console_scripts': [
              'openapi-type = openapi_type.cli:main'
          ],
      }
    )



import argparse
import json
import sys
from pathlib import Path
from typing import Mapping
from itertools import islice

from openapi_type import parse_spec


def is_empty_dir(p: Path) -> bool:
    return p.is_dir() and not bool(list(islice(p.iterdir(), 1)))


def setup(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    sub = subparsers.add_parser('check', help='Check whether a provided schema (JSON, YAML) can be parsed.')
    sub.add_argument('-s', '--source', help="Path to a spec (JSON, YAML). "
                                            "If not specified, then the data will be read from stdin.")
    sub.set_defaults(run_cmd=main)
    return sub


def main(args: argparse.Namespace, in_channel=sys.stdin, out_channel=sys.stdout) -> None:
    """ $ <cmd-prefix> gen <source> <target>
    """
    try:
        with Path(args.source).open('r') as f:
            python_data = _read_data(f)
    except TypeError:
        # source is None, read from stdin
        python_data = _read_data(in_channel)

    _spec = parse_spec(python_data)

    out_channel.write('Successfully parsed.\n')


def _read_data(fd) -> Mapping:
    buf = fd.read()  # because stdin does not support seek and we want to try both json and yaml parsing
    try:
        struct = json.loads(buf)
    except ValueError:
        try:
            import yaml
        except ImportError:
            raise RuntimeError(
                "Could not parse data as JSON, and could not locate PyYAML library "
                "to try to parse the data as YAML. You can either install PyYAML as a separate "
                "dependency, or use the `third_party` extra tag:\n\n"
                "$ pip install openapi-client-generator[third_party]"
            )
        struct = yaml.full_load(buf)
    return struct
